/*******************************************************************\

Module: Dependencies between events for partial order concurrency

Author: Michael Tautschnig, michael.tautschnig@cs.ox.ac.uk

\*******************************************************************/

#include <cassert>
#include <set>
#include <map>
#include <sstream>

#include <std_expr.h>
#include <message.h>

#include <goto-programs/goto_program.h>

#define DUMMY_FOR_SEMICOLON ((void)0)
#define DP_DEBUG(name, e1, e2) \
  { \
    std::ostringstream oss; \
    oss << "DP " << name << " " << e1 << " -> " << e2; \
    message.debug(oss.str()); \
  } DUMMY_FOR_SEMICOLON

/*******************************************************************\

Function: collect_symbols

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

static void collect_symbols(
    const exprt &expr,
    std::set<exprt const *> &symbols)
{
	if (expr.id()==ID_symbol)
		symbols.insert(&expr);
	else {
		forall_operands(iter, expr)
      collect_symbols(*iter, symbols);
	}
}

/*******************************************************************\

Function: collect_expr_with_parents

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

static void collect_expr_with_parents(
    const exprt &expr,
    std::map<exprt const*, exprt const*> &exprs)
{
	exprs.insert(
      std::make_pair<exprt const*, exprt const*>(&expr, 0));
	forall_operands(iter, expr) {
			exprs.insert(std::make_pair(&(*iter), &expr));
			collect_expr_with_parents(*iter, exprs);
	}
}

/*******************************************************************\

Function: collect_parents

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

static void collect_parents(
    const symbol_exprt &sym,
    goto_programt::const_targett pc,
    std::set<exprt const*> &parents)
{
  std::map<exprt const*, exprt const*> subexprt_to_parents;
  collect_expr_with_parents(pc->code, subexprt_to_parents);
  std::set<exprt const*> contained_symbols;
  collect_symbols(pc->code, contained_symbols);

  std::set<exprt const*> dereferenced;
  for(std::set<exprt const*>::const_iterator
      it=contained_symbols.begin();
      it!=contained_symbols.end();
      ++it)
  {
    const irep_idt &identifier=to_symbol_expr(**it).get_identifier();
    for(exprt const* parent=*it; parent!=0; )
    {
      if(identifier==sym.get_identifier())
        parents.insert(parent);
      else if(parent->id()==ID_dereference)
        dereferenced.insert(parent);

      std::map<exprt const*, exprt const*>::const_iterator
        entry=subexprt_to_parents.find(parent);
      assert(entry!=subexprt_to_parents.end());
      parent=entry->second;
    }
  }
  if(parents.empty())
    parents.swap(dereferenced);
}

/*******************************************************************\

Function: get_depended_on

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

static void get_depended_on(
    const symbol_exprt &sym,
    goto_programt::const_targett pc,
    const bool with_addr_deps,
    const bool with_data_deps,
    std::set<exprt const *> &deps,
    std::set<irep_idt> &killed)
{
  std::set<exprt const*> parents;
  collect_parents(sym, pc, parents);

  std::set<exprt const*> lhs;
  for(std::set<exprt const*>::const_iterator it=parents.begin();
      it!=parents.end();
      ++it)
  {
    if(with_addr_deps && (*it)->id()==ID_index)
      // it might be necessary to check whether e1 occurs as array and e2 as
      // index; we skip this for now
      collect_symbols(**it, deps);
    else if(with_addr_deps && (*it)->id()==ID_dereference)
      collect_symbols(**it, deps);
    else if(with_data_deps && ((*it)->id()==ID_assign ||
          (*it)->id()==ID_code && to_code(**it).get_statement()==ID_assign) &&
        parents.find(&((*it)->op0()))!=parents.end())
    {
      collect_symbols((*it)->op1(), deps);
      collect_symbols((*it)->op0(), lhs);
    }
  }
  for(std::set<exprt const*>::const_iterator it=lhs.begin();
      it!=lhs.end();
      ++it)
    killed.insert(to_symbol_expr(**it).get_identifier());
  for(std::set<exprt const*>::const_iterator it=deps.begin();
      it!=deps.end();
      ++it)
  {
    std::set<irep_idt>::iterator k_entry=
      killed.find(to_symbol_expr(**it).get_identifier());
    if(k_entry!=killed.end())
      killed.erase(k_entry);
  }
}

/*******************************************************************\

Function: get_dependencies_at

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

static void get_dependencies_at(
    const symbol_exprt &e2_sym,
    goto_programt::const_targett e2_pc,
    goto_programt::const_targett e1_pc,
    const bool dp_addr_only,
    const bool recurse,
    const namespacet &ns,
    std::set<exprt const *> &deps_at_pc)
{
  assert(deps_at_pc.empty());

  // we only do function-local dependencies -- may be an overapproximation
  if(e1_pc->function!=e2_pc->function) return;

  // make sure we can linearly get from e1's pc to e2's pc -- another
  // overapproximation
  if(!(e1_pc/*->variable_number*/<e2_pc/*->variable_number*/)) return;

  // collect dependencies of e2_sym at e2_pc and then recursively collect
  // dependencies live at e1_pc
  std::set<exprt const*> all_deps;
  std::set<irep_idt> killed_at_pc;
  get_depended_on(e2_sym, e2_pc, true, !dp_addr_only,
      deps_at_pc, killed_at_pc);

  for(goto_programt::const_targett it=e2_pc;
      it!=e1_pc;
      --it)
  {
    assert(!(it/*->variable_number*/<e1_pc/*->variable_number*/));

    goto_programt::const_targett pred=it;
    --pred;

    all_deps.insert(deps_at_pc.begin(), deps_at_pc.end());
    deps_at_pc.clear();
    std::set<irep_idt> killed_at_pc_bak;
    killed_at_pc_bak.swap(killed_at_pc);
    for(std::set<exprt const*>::iterator it_d=all_deps.begin();
        it_d!=all_deps.end();
        ) // no ++it_d
    {
      const symbol_exprt &sym=to_symbol_expr(**it_d);
      if(killed_at_pc_bak.find(sym.get_identifier())!=
          killed_at_pc_bak.end() ||
          (!recurse && is_global(ns.lookup(sym.get_identifier()))))
      {
        std::set<exprt const*>::iterator next=it_d;
        ++next;
        all_deps.erase(it_d);
        it_d=next;
      }
      else
      {
        ++it_d;
        get_depended_on(sym, pred, false, true, deps_at_pc, killed_at_pc);
      }
    }
  }
}
